//    Copyright (C) 2006-2007  Regis COSNIER
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;
using System.Windows.Forms;

using System.Threading;
using System.Windows.Threading;
using System.Windows.Media;
using System.Windows.Media.Imaging;

using Microsoft.Win32;

using DirectShowLib;
using DirectShowLib.BDA;
using DirectShowLib.Utils;

namespace CodeTV
{
	public class GraphBuilderBase : IDisposable, VideoControl.IVideoEventHandler, ISampleGrabberCB
    {
		public static TraceSwitch trace = new TraceSwitch("GraphBuilder", "GraphBuilder traces", "Info");

		protected Settings settings;

		protected VideoControl hostingControl;
		protected bool useWPF = true;
		//protected DirectDraw directDraw = new DirectDraw();

		protected static Dictionary<string, DsDevice> audioRendererDevices;
		protected DsDevice audioRendererDevice;

		protected IFilterGraph2 graphBuilder;
		protected DsROTEntry rot;

		protected IBaseFilter audioRenderer;
		protected IBaseFilter videoRenderer;

		protected Size currentVideoSourceSize;
		protected Rectangle currentVideoTargetRectangle;
		protected VideoSizeMode videoZoomMode = VideoSizeMode.FromInside;
		protected bool videoKeepAspectRatio = true;
		protected PointF videoOffset = new PointF(0.5f, 0.5f);
		protected double videoZoom = 1.0;
		protected double videoAspectRatioFactor = 1.0;
		protected bool useVideo169Mode = false;

		public Settings Settings { get { return this.settings; } set { this.settings = value; } }

		public static Dictionary<string, DsDevice> AudioRendererDevices
		{
			get
			{
				if (audioRendererDevices == null)
				{
					audioRendererDevices = new Dictionary<string, DsDevice>();

					DsDevice[] devices = DsDevice.GetDevicesOfCat(FilterCategory.AudioRendererCategory);
					foreach (DsDevice d in devices)
						audioRendererDevices.Add(d.Name, d);
				}
				return audioRendererDevices;
			}
		}

		public DsDevice AudioRendererDevice { get { return this.audioRendererDevice; } set { this.audioRendererDevice = value; } }

        public virtual IFilterGraph2 FilterGraph { get { return this.graphBuilder; } }
		public IBaseFilter AudioRenderer { get { return this.audioRenderer; } }
		public IBaseFilter VideoRenderer { get { return this.videoRenderer; } }

		public Size CurrentVideoSourceSize { get { return this.currentVideoSourceSize; } set { this.currentVideoSourceSize = value; } }
		public Rectangle CurrentVideoTargetRectangle { get { return this.currentVideoTargetRectangle; } set { this.currentVideoTargetRectangle = value; } }
		public VideoSizeMode VideoZoomMode { get { return this.videoZoomMode; } set { this.videoZoomMode = value; } }
		public bool VideoKeepAspectRatio { get { return this.videoKeepAspectRatio; } set { this.videoKeepAspectRatio = value; } }
		public PointF VideoOffset { get { return this.videoOffset; } set { this.videoOffset = value; } }
		public double VideoZoom { get { return this.videoZoom; } set { this.videoZoom = value; } }
		public double VideoAspectRatioFactor { get { return this.videoAspectRatioFactor; } set { this.videoAspectRatioFactor = value; } }
		public bool UseVideo169Mode { get { return this.useVideo169Mode; } set { this.useVideo169Mode = value; } }

		public event EventHandler GraphStarted;
		public event EventHandler GraphEnded;
		public event EventHandler<PossibleEventArgs> PossibleChanged;

		public class PossibleEventArgs: EventArgs
		{
			public PossibleEventArgs(string possible, bool isPossible) { this.possible = possible; this.isPossible = isPossible; }
			public string possible;
			public bool isPossible;
		}

		protected void OnPossibleChanged(string possible, bool isPossible)
		{
			if (PossibleChanged != null)
				PossibleChanged(this, new PossibleEventArgs(possible, isPossible));
		}

		protected bool isPossibleGraphRun = false;
		protected bool isPossibleGraphPause = false;
		protected bool isPossibleGraphStop = false;
		protected bool isPossibleGraphRestart = false;
		protected bool isPossibleGraphRelease = false;

		protected bool isPossibleSetSpeed = false;
		protected bool isPossibleSetPosition = false;

		public bool IsPossibleGraphRun { get { return this.isPossibleGraphRun; } protected set { if (this.isPossibleGraphRun != value) { this.isPossibleGraphRun = value; OnPossibleChanged("IsPossibleGraphRun", value); } } }
		public bool IsPossibleGraphPause { get { return this.isPossibleGraphPause; } protected set { if (this.isPossibleGraphPause != value) { this.isPossibleGraphPause = value; OnPossibleChanged("IsPossibleGraphPause", value); } } }
		public bool IsPossibleGraphStop { get { return this.isPossibleGraphStop; } protected set { if (this.isPossibleGraphStop != value) { this.isPossibleGraphStop = value; OnPossibleChanged("IsPossibleGraphStop", value); } } }
		public bool IsPossibleGraphRestart { get { return this.isPossibleGraphRestart; } protected set { if (this.isPossibleGraphRestart != value) { this.isPossibleGraphRestart = value; OnPossibleChanged("IsPossibleGraphRestart", value); } } }
		public bool IsPossibleGraphRelease { get { return this.isPossibleGraphRelease; } protected set { if (this.isPossibleGraphRelease != value) { this.isPossibleGraphRelease = value; OnPossibleChanged("IsPossibleGraphRelease", value); } } }

		public bool IsPossibleSetSpeed { get { return this.isPossibleSetSpeed; } protected set { if (this.isPossibleSetSpeed != value) { this.isPossibleSetSpeed = value; OnPossibleChanged("IsPossibleSetSpeed", value); } } }
		public bool IsPossibleSetPosition { get { return this.isPossibleSetPosition; } protected set { if (this.isPossibleSetPosition != value) { this.isPossibleSetPosition = value; OnPossibleChanged("IsPossibleSetPosition", value); } } }


		public GraphBuilderBase(VideoControl renderingControl)
        {
            this.hostingControl = renderingControl;

			//directDraw.Init();
        }

		public virtual void BuildGraph()
		{
			useWPF = Settings.UseWPF;

			this.graphBuilder = (IFilterGraph2)new FilterGraph();
			rot = new DsROTEntry(this.graphBuilder);

			AddRenderers();
			if(!useWPF)
				ConfigureVMR9InWindowlessMode();

			this.hostingControl.CurrentGraphBuilder = this;

			OnGraphStarted();
		}

		public virtual void OnGraphStarted()
		{
			if (GraphStarted != null)
				GraphStarted(this, new EventArgs());

			IsPossibleGraphRun = true;
			IsPossibleGraphPause = false;
			IsPossibleGraphStop = true;
			IsPossibleGraphRestart = false;
			IsPossibleGraphRelease = true;
		}

		public virtual void OnGraphEnded()
		{
			IsPossibleGraphRun = false;
			IsPossibleGraphPause = false;
			IsPossibleGraphStop = false;
			IsPossibleGraphRestart = false;
			IsPossibleGraphRelease = false;

			if (GraphEnded != null)
				GraphEnded(this, new EventArgs());
		}

		public virtual FilterState GetGraphState()
		{
			IMediaControl mediaControl = this.graphBuilder as IMediaControl;
			FilterState pfs;
			mediaControl.GetState(0, out pfs);
			return pfs;
		}

		public virtual void RunGraph()
		{
			IMediaControl mediaControl = this.graphBuilder as IMediaControl;
			FilterState pfs;
			mediaControl.GetState(0, out pfs);
			if (pfs != FilterState.Running)
			{
				int hr = mediaControl.Run();
				ThrowExceptionForHR("Running the graph: ", hr);

				IsPossibleGraphRun = false;
				IsPossibleGraphPause = true;
				IsPossibleGraphStop = true;
			}
		}

		public virtual void PauseGraph()
		{
			IMediaControl mediaControl = this.graphBuilder as IMediaControl;
			FilterState pfs;
			mediaControl.GetState(0, out pfs);
			if (pfs == FilterState.Running)
			{
				int hr = mediaControl.Pause();
				ThrowExceptionForHR("Pausing the graph: ", hr);

				IsPossibleGraphRun = true;
				IsPossibleGraphPause = false;
				IsPossibleGraphStop = true;
			}
		}

		public virtual void StopGraph()
		{
			IMediaControl mediaControl = this.graphBuilder as IMediaControl;
			FilterState pfs;
			mediaControl.GetState(0, out pfs);
			if (pfs == FilterState.Running || pfs == FilterState.Paused)
			{
				int hr = mediaControl.Stop();
				ThrowExceptionForHR("Stopping the graph: ", hr);

				IsPossibleGraphRun = true;
				IsPossibleGraphPause = false;
				IsPossibleGraphStop = false;
			}
		}

		public virtual void SaveGraph(string filepath)
        {
            FilterGraphTools.SaveGraphFile(this.graphBuilder, filepath);
        }

		public virtual void OnVideoEvent(int cookies)
		{
			if (this.graphBuilder == null)
				return;

			IMediaEventEx mediaEvent = this.graphBuilder as IMediaEventEx;
			EventCode eventCode;
			IntPtr param1, param2;
			while (mediaEvent.GetEvent(out eventCode, out param1, out param2, 0) >= 0)
			{
				Trace.WriteLineIf(trace.TraceVerbose, "OnVideoEvent() -> " + eventCode.ToString());

				if (eventCode == EventCode.VMRRenderDeviceSet)
					VideoRefresh();
				//if (eventCode == EventCode.Paused || eventCode == EventCode.NeedRestart)
				//{
				//    RunGraph();
				//}
				//switch (eventCode) 
				//{ 
				//    // Call application-defined functions for each 
				//    // type of event that you want to handle.
				//}
				int hr = mediaEvent.FreeEventParams(eventCode, param1, param2);
			}
		}

        #region Membres de IDisposable

		public virtual void Dispose()
        {
            Decompose();
        }

        #endregion


		protected virtual void AddRenderers()
		{
			AddRenderers(this.graphBuilder);
		}

		protected virtual void AddRenderers(IFilterGraph2 graphBuilder)
		{
			int hr = 0;
			Guid iid = typeof(IBaseFilter).GUID;

			if (this.AudioRendererDevice != null)
			{
				try
				{
					this.audioRenderer = null;

					IBaseFilter tmp;

					object o;
					this.AudioRendererDevice.Mon.BindToObject(null, null, ref iid, out o);
					tmp = o as IBaseFilter;

					//Add the Video input device to the graph
					hr = graphBuilder.AddFilter(tmp, this.AudioRendererDevice.Name);
					if (hr >= 0)
					{
						// Got it !
						this.audioRenderer = tmp;
					}
					else
					{
						// Try another...
						int hr1 = graphBuilder.RemoveFilter(tmp);
						Marshal.ReleaseComObject(tmp);
						//DsError.ThrowExceptionForHR(hr);
					}
				}
				catch { }
			}

			if (this.audioRenderer == null)
			{
				// Add default audio renderer
				this.audioRenderer = (IBaseFilter)new DSoundRender();
				hr = graphBuilder.AddFilter(this.audioRenderer, "DirectSound Renderer");
				ThrowExceptionForHR("Adding the DirectSound Renderer: ", hr);
			}

			// To see something

			if (useWPF)
			{
				// Get the SampleGrabber interface
				ISampleGrabber sampleGrabber = new SampleGrabber() as ISampleGrabber;
				this.videoRenderer = sampleGrabber as IBaseFilter;

				// Set the media type to Video
				AMMediaType media = new AMMediaType();
				media.majorType = MediaType.Video;
				//media.subType = MediaSubType.YUY2; // RGB24;
				//media.formatType = FormatType.Null;
				//media.sampleSize = 1;
				//media.temporalCompression = false;
				//media.fixedSizeSamples = false;
				//media.unkPtr = IntPtr.Zero;
				//media.formatType = FormatType.None;
				//media.formatSize = 0;
				//media.formatPtr = IntPtr.Zero;
				media.subType = MediaSubType.RGB32; // RGB24;
				media.formatType = FormatType.VideoInfo;
				hr = sampleGrabber.SetMediaType(media);
				ThrowExceptionForHR("Setting the MediaType on the SampleGrabber: ", hr);
				DsUtils.FreeAMMediaType(media);

				// Configure the samplegrabber
				hr = sampleGrabber.SetCallback(this, 1);
				DsError.ThrowExceptionForHR(hr);


				// Add the frame grabber to the graph
				hr = graphBuilder.AddFilter(this.videoRenderer, "SampleGrabber");
				ThrowExceptionForHR("Adding the SampleGrabber: ", hr);

				//hr = ConnectFilters(this.videoRenderer, nullRenderer);
			}
			else
			{
				this.videoRenderer = (IBaseFilter)new VideoMixingRenderer9();
				hr = graphBuilder.AddFilter(this.videoRenderer, "Video Mixing Renderer Filter 9");
				ThrowExceptionForHR("Adding the VMR9: ", hr);
			}
			//IReferenceClock clock = this.audioRenderer as IReferenceClock;
			//if(clock != null)
			//{
			//    // Set the graph clock.
			//    this.FilterGraph.SetDefaultSyncSource( 
			//    pGraph->QueryInterface(IID_IMediaFilter, (void**)&pMediaFilter);
			//    pMediaFilter->SetSyncSource(pClock);
			//    pClock->Release();
			//    pMediaFilter->Release();
			//}
		}

		protected void AddAndConnectNullRendererForWPF()
		{
			if (useWPF)
			{
				// In order to keep the audio/video synchro, we need the NullRenderer
				IBaseFilter nullRenderer = new NullRenderer() as IBaseFilter;
				int hr = graphBuilder.AddFilter(nullRenderer, "NullRenderer");
				ThrowExceptionForHR("Adding the NullRenderer: ", hr);

				IPin pinOutFromFilterOut = DsFindPin.ByDirection(this.videoRenderer, PinDirection.Output, 0);
				if (pinOutFromFilterOut != null)
				{
					try
					{
						IPin pinInFromFilterOut = DsFindPin.ByDirection(nullRenderer, PinDirection.Input, 0);
						if (pinInFromFilterOut != null)
						{
							try
							{
								hr = this.graphBuilder.Connect(pinOutFromFilterOut, pinInFromFilterOut);
							}
							finally
							{
								Marshal.ReleaseComObject(pinInFromFilterOut);
							}
						}
					}
					finally
					{
						Marshal.ReleaseComObject(pinOutFromFilterOut);
					}
				}

				//IPin pinOutFromFilterOut = DsFindPin.ByDirection(this.videoRenderer, PinDirection.Output, 0);
				//if (pinOutFromFilterOut != null)
				//{
				//    hr = this.graphBuilder.Render(pinOutFromFilterOut);
				//    Marshal.ReleaseComObject(pinOutFromFilterOut);
				//}
			}
		}

		protected void ThrowExceptionForHR(string errorText, int hr)
		{
            // If a severe error has occurred
            if (hr < 0)
            {

				try
				{
					string s = DsError.GetErrorText(hr);

					// If a string is returned, build a com error from it
					if (s != null)
					{
						throw new COMException(s, hr);
					}
					else
					{
						// No string, just use standard com error
						Marshal.ThrowExceptionForHR(hr);
					}
				}
				catch (Exception ex)
				{
					throw new Exception(string.Format("{0}{1}", errorText, ex.Message), ex);
				}
			}
		}

		protected int ConnectFilters(IBaseFilter filterIn, IBaseFilter filterOut)
		{
			int hr = -1;
			IPin pinOutFromFilterIn = DsFindPin.ByDirection(filterIn, PinDirection.Output, 0);
			if (pinOutFromFilterIn != null)
			{
				IPin pinInFromFilterOut = DsFindPin.ByDirection(filterOut, PinDirection.Input, 0);
				if (pinInFromFilterOut != null)
				{
					hr = this.graphBuilder.Connect(pinOutFromFilterIn, pinInFromFilterOut);
					Marshal.ReleaseComObject(pinInFromFilterOut);
				}

				Marshal.ReleaseComObject(pinOutFromFilterIn);
			}

			return hr;
		}

		//private IntPtr m_handle = IntPtr.Zero;
		//private int m_videoWidth;
		//private int m_videoHeight;
		//private int m_stride;

		protected void WpfUpdateVideoSize()
		{
			if (this.videoRenderer is ISampleGrabber)
			{
				try
				{
					ISampleGrabber sampGrabber = this.videoRenderer as ISampleGrabber;

					// Get the media type from the SampleGrabber
					AMMediaType media = new AMMediaType();
					int hr = sampGrabber.GetConnectedMediaType(media);
					DsError.ThrowExceptionForHR(hr);

					if ((media.formatType != FormatType.VideoInfo) || (media.formatPtr == IntPtr.Zero))
					{
						throw new NotSupportedException("Unknown Grabber Media Format");
					}


					// Grab the size info
					VideoInfoHeader videoInfoHeader = (VideoInfoHeader)Marshal.PtrToStructure(media.formatPtr, typeof(VideoInfoHeader));
					int videoWidth = videoInfoHeader.BmiHeader.Width;
					int videoHeight = videoInfoHeader.BmiHeader.Height;
					int stride = videoWidth * (videoInfoHeader.BmiHeader.BitCount / 8);

					DsUtils.FreeAMMediaType(media);
					media = null;


					int bytePerPixel = 4;
					// These are 'dummy' pixels only used to create our bitmap
					// further editing after bitmap creation of these pixels does nothing,
					// that is what this hack is for
					byte[] frameBuffer = new byte[videoWidth * videoHeight * bytePerPixel];
					//Create a new bitmap source
					BitmapSource bitmapSource = BitmapSource.Create(videoWidth,
															  videoHeight,
															  96,
															  96,
															  PixelFormats.Bgr32,
															  null,
															  frameBuffer,
															  videoWidth * bytePerPixel);

					this.hostingControl.WPFImage.Source = bitmapSource;
					lock (this.hostingControl.WPFImage)
					{
						this.wpfBitmapBuffer = new WPFUtil.BitmapBuffer(bitmapSource);
					}
				}
				catch (Exception)
				{
				}
			}
		}

		protected void WPFStop()
		{
			if (this.hostingControl != null && this.hostingControl.WPFImage != null)
			{
				lock (this.hostingControl.WPFImage)
				{
					this.wpfBitmapBuffer = null;
				}
			}
		}

		WPFUtil.BitmapBuffer wpfBitmapBuffer;
		[DllImport("Kernel32.dll", EntryPoint = "RtlMoveMemory")]
		private static extern void CopyMemory(IntPtr Destination, IntPtr Source, int Length);

		/// <summary> sample callback, NOT USED. </summary>
		int ISampleGrabberCB.SampleCB(double sampleTime, IMediaSample mediaSample)
		{
			return 0;
		}

		/// <summary> buffer callback, COULD BE FROM FOREIGN THREAD. </summary>
		unsafe int ISampleGrabberCB.BufferCB(double sampleTime, IntPtr buffer, int bufferLength)
		{
			System.Windows.Controls.Image image = this.hostingControl.WPFImage;
			if (image != null)
			{
				image.Dispatcher.BeginInvoke(DispatcherPriority.Render, new ThreadStart(delegate
				{
					//lock (image)
					{
						if (wpfBitmapBuffer != null)
						{
							// TODO: synchronisation between each frame here and the monitor VBL
							//directDraw.WaitForVerticalBlank();
							CopyMemory(wpfBitmapBuffer.BufferPointer, buffer, bufferLength);
							image.InvalidateVisual();
						}
					}
				}));
			}

			return 0;
		}

		protected virtual void ConfigureVMR9InWindowlessMode()
		{
			ConfigureVMR9InWindowlessMode(1);
		}

		protected virtual void ConfigureVMR9InWindowlessMode(int numberOfStream)
        {
            int hr = 0;
            IVMRFilterConfig9 filterConfig = this.videoRenderer as IVMRFilterConfig9;

            // Configure VMR-9 in Windowsless mode
			hr = filterConfig.SetRenderingMode(VMR9Mode.Windowless);
			//hr = filterConfig.SetRenderingMode(VMR9Mode.Windowed);
			ThrowExceptionForHR("Setting the VMR9 RenderingMode: ", hr);

            // With 1 input stream
			hr = filterConfig.SetNumberOfStreams(numberOfStream);
			ThrowExceptionForHR("Setting the VMR9 NumberOfStreams: ", hr);

            IVMRWindowlessControl9 windowlessControl = this.videoRenderer as IVMRWindowlessControl9;

            // The main form is hosting the VMR-9
            hr = windowlessControl.SetVideoClippingWindow(this.hostingControl.Handle);
			ThrowExceptionForHR("Setting the VMR9 VideoClippingWindow: ", hr);

            // Keep the aspect-ratio OK
			//hr = windowlessControl.SetAspectRatioMode(VMR9AspectRatioMode.LetterBox);
			hr = windowlessControl.SetAspectRatioMode(VMR9AspectRatioMode.None);
			ThrowExceptionForHR("Setting the VMR9 AspectRatioMode: ", hr);

            // Init the VMR-9 with default size values
            OnResizeMoveHandler(null, null);

            // Add Windows Messages handlers
            AddHandlers();
		}


		//private Color colorKey = Color.Violet; // The color use as ColorKey for GDI operations
		private System.Drawing.Color colorKey = System.Drawing.Color.Violet; // The color use as ColorKey for GDI operations
		private Bitmap colorKeyBitmap; // A RGB bitmap used for GDI operations.
		//private Bitmap alphaBitmap; // A ARGB bitmap used for Direct3D operations

		//// Managed Direct3D magic number to retrieve unmanaged Direct3D interfaces
		//private const int DxMagicNumber = -759872593;
		//private Device device = null; // A Managed Direct3D device
		//private PresentParameters presentParams;
		//private Surface surface = null; // A Direct3D suface filled with alphaBitmap
		//private IntPtr unmanagedSurface; // A pointer on the unmanaged surface

		//private void InitializeDirect3D()
		//{
		//    // Basic Presentation Parameters...
		//    PresentParameters presentParams = new PresentParameters();
		//    presentParams.Windowed = true;
		//    presentParams.SwapEffect = SwapEffect.Discard;

		//    // Assume a hardware Direct3D device is available
		//    // Add MultiThreaded to be safe. Each DirectShow filter runs in a separate thread...
		//    device = new Device(
		//        0,
		//        DeviceType.Hardware,
		//        this,
		//        CreateFlags.SoftwareVertexProcessing | CreateFlags.MultiThreaded,
		//        presentParams
		//        );

		//    // Create a surface from our alpha bitmap
		//    surface = new Surface(device, alphaBitmap, Pool.SystemMemory);
		//    // Get the unmanaged pointer
		//    unmanagedSurface = surface.GetObjectByValue(DxMagicNumber);
		//}

		//protected override void Dispose(bool disposing)
		//{

		//    // Dispose Managed Direct3D objects
		//    surface.Dispose();
		//    device.Dispose();

		//    base.Dispose(disposing);
		//}

		private NormalizedRect GetDestRectangle()
		{
			int hr = 0;
			int width, height, arW, arH;
			NormalizedRect rect = new NormalizedRect();

			hr = (this.videoRenderer as IVMRWindowlessControl9).GetNativeVideoSize(out width, out height, out arW, out arH);
			DsError.ThrowExceptionForHR(hr);

			// Position the bitmap in the middle of the video stream.
			if (width >= height)
			{
				rect.top = 0.0f;
				rect.left = (1.0f - ((float)height / (float)width)) / 2;
				rect.bottom = 1.0f;
				rect.right = rect.left + (float)height / (float)width;
			}
			else
			{
				rect.top = (1.0f - ((float)width / (float)height)) / 2;
				rect.left = 0.0f;
				rect.right = rect.top + (float)width / (float)height;
				rect.bottom = 1.0f;
			}

			return rect;
		}

		public void StartOSD()
		{
			// Get the colorkeyed bitmap without antialiasing
			colorKeyBitmap = BitmapGenerator.GenerateColorKeyBitmap(colorKey, false);
			//// Get the bitmap with alpha transparency
			//alphaBitmap = BitmapGenerator.GenerateAlphaBitmap();


			IVMRMixerBitmap9 mixerBitmap = this.videoRenderer as IVMRMixerBitmap9;

			//if (usingGDI)
			//{
				// Old school GDI stuff...
				Graphics g = Graphics.FromImage(colorKeyBitmap);
				IntPtr hdc = g.GetHdc();
				IntPtr memDC = NativeMethodes.CreateCompatibleDC(hdc);
				IntPtr hBitmap = colorKeyBitmap.GetHbitmap();
				NativeMethodes.SelectObject(memDC, hBitmap);

				// Set Alpha Bitmap Parameters for using a GDI DC
				VMR9AlphaBitmap alphaBmp = new VMR9AlphaBitmap();
				alphaBmp.dwFlags = VMR9AlphaBitmapFlags.hDC | VMR9AlphaBitmapFlags.SrcColorKey | VMR9AlphaBitmapFlags.FilterMode;
				alphaBmp.hdc = memDC;
				alphaBmp.rSrc = new DsRect(0, 0, colorKeyBitmap.Size.Width, colorKeyBitmap.Size.Height);
				alphaBmp.rDest = GetDestRectangle();
				alphaBmp.clrSrcKey = ColorTranslator.ToWin32(colorKey);
				alphaBmp.dwFilterMode = VMRMixerPrefs.PointFiltering;
				alphaBmp.fAlpha = 0.75f;

				// Set Alpha Bitmap Parameters
				int hr = mixerBitmap.SetAlphaBitmap(ref alphaBmp);
				DsError.ThrowExceptionForHR(hr);

				// Release GDI handles
				NativeMethodes.DeleteObject(hBitmap);
				NativeMethodes.DeleteDC(memDC);
				g.ReleaseHdc(hdc);
				g.Dispose();
			//}
			//else // Using a Direct3D surface
			//{
			//    // Set Alpha Bitmap Parameters for using a Direct3D surface
			//    VMR9AlphaBitmap alphaBmp = new VMR9AlphaBitmap();
			//    alphaBmp.dwFlags = VMR9AlphaBitmapFlags.EntireDDS;
			//    alphaBmp.pDDS = unmanagedSurface;
			//    alphaBmp.rDest = GetDestRectangle();
			//    alphaBmp.fAlpha = 1.0f;
			//    // Note : Alpha values from the bitmap are cumulative with the fAlpha parameter.
			//    // Example : texel alpha = 128 (50%) & fAlpha = 0.5f (50%) = effective alpha : 64 (25%)

			//    // Set Alpha Bitmap Parameters
			//    int hr = mixerBitmap.SetAlphaBitmap(ref alphaBmp);
			//    DsError.ThrowExceptionForHR(hr);
			//}
		}

		public void StopOSD()
		{
			IVMRMixerBitmap9 mixerBitmap = this.videoRenderer as IVMRMixerBitmap9;

			// Get current Alpha Bitmap Parameters
			VMR9AlphaBitmap alphaBmp;
			int hr = mixerBitmap.GetAlphaBitmapParameters(out alphaBmp);
			DsError.ThrowExceptionForHR(hr);

			// Disable them
			alphaBmp.dwFlags = VMR9AlphaBitmapFlags.Disable;

			// Update the Alpha Bitmap Parameters
			hr = mixerBitmap.UpdateAlphaBitmapParameters(ref alphaBmp);
			DsError.ThrowExceptionForHR(hr);
		}

		protected virtual void ConnectAllOutputFiltersFrom(IBaseFilter fromFilter, IFilterGraph2 graph)
		{
			// Get the pin enumerator
			IEnumPins ppEnum;
			int hr = fromFilter.EnumPins(out ppEnum);
			DsError.ThrowExceptionForHR(hr);

			try
			{
				// Walk the pins looking for a match
				IPin[] pPins = new IPin[1];
				//22 int lFetched;
				//22 while ((ppEnum.Next(1, pPins, out lFetched) >= 0) && (lFetched == 1))
				while (ppEnum.Next(1, pPins, IntPtr.Zero) >= 0)
				{
					// Read the direction
					PinDirection ppindir;
					hr = pPins[0].QueryDirection(out ppindir);
					DsError.ThrowExceptionForHR(hr);

					// Is it the right direction?
					if (ppindir == PinDirection.Output)
					{
						if (pPins[0] != null)
						{
							hr = graph.Render(pPins[0]);
							//DsError.ThrowExceptionForHR(hr);
						}
					}
					Marshal.ReleaseComObject(pPins[0]);
				}
			}
			finally
			{
				Marshal.ReleaseComObject(ppEnum);
			}
		}

		/// <summary>
		/// Enumerates all filters of the selected category and returns the IBaseFilter for the 
		/// filter described in friendlyname
		/// </summary>
		/// <param name="category">Category of the filter</param>
		/// <param name="friendlyname">Friendly name of the filter</param>
		/// <returns>IBaseFilter for the device</returns>
		protected virtual IBaseFilter CreateFilter(Guid category, string friendlyname)
		{
			object source = null;
			Guid iid = typeof(IBaseFilter).GUID;
			foreach (DsDevice device in DsDevice.GetDevicesOfCat(category))
			{
				if (device.Name.CompareTo(friendlyname) == 0)
				{
					device.Mon.BindToObject(null, null, ref iid, out source);
					break;
				}
			}

			return (IBaseFilter)source;
		}

		protected virtual void AddHandlers()
        {
            // Add Windows Messages handlers
			this.hostingControl.PaintBackground += new VideoControl.PaintBackgroundEventHandler(OnPaintBackground);
			this.hostingControl.Paint += new PaintEventHandler(OnPaintHandler); // for WM_PAINT
			this.hostingControl.Resize += new EventHandler(OnResizeMoveHandler); // for WM_SIZE
            //this.hostingControl.Move += new EventHandler(OnResizeMoveHandler); // for WM_MOVE
            SystemEvents.DisplaySettingsChanged += new EventHandler(OnDisplayChangedHandler); // for WM_DISPLAYCHANGE
        }

		protected virtual void RemoveHandlers()
        {
            // Remove Windows Messages handlers
			this.hostingControl.PaintBackground -= new VideoControl.PaintBackgroundEventHandler(OnPaintBackground);
			this.hostingControl.Paint -= new PaintEventHandler(OnPaintHandler); // for WM_PAINT
            this.hostingControl.Resize -= new EventHandler(OnResizeMoveHandler); // for WM_SIZE
            //this.hostingControl.Move -= new EventHandler(OnResizeMoveHandler); // for WM_MOVE
            SystemEvents.DisplaySettingsChanged -= new EventHandler(OnDisplayChangedHandler); // for WM_DISPLAYCHANGE
        }

		protected virtual void Decompose()
        {
			if (this.graphBuilder != null)
			{
				int hr = 0;

				OnGraphEnded();

				// Decompose the graph
				try { hr = (this.graphBuilder as IMediaControl).StopWhenReady(); }
				catch { }
				try { hr = (this.graphBuilder as IMediaControl).Stop(); }
				catch { }
				RemoveHandlers();

				FilterGraphTools.RemoveAllFilters(this.graphBuilder);

				if (this.audioRenderer != null) Marshal.ReleaseComObject(this.audioRenderer); this.audioRenderer = null;
				if (this.videoRenderer != null) Marshal.ReleaseComObject(this.videoRenderer); this.videoRenderer = null;

				try { rot.Dispose(); }
				catch { }
				try { Marshal.ReleaseComObject(this.graphBuilder); this.graphBuilder = null; }
				catch { }
			}
        }

		//protected virtual Rectangle GetInnerRectangle(Rectangle outerRectangle, double innerRatio)
		//{
		//    Rectangle innerRectangle = outerRectangle;
		//    double outerRatio = (double)outerRectangle.Width / (double)outerRectangle.Height;
		//    if (outerRatio >= innerRatio)
		//    {
		//        innerRectangle.Width = (int)((double)outerRectangle.Height * innerRatio);
		//        innerRectangle.X = (int)((outerRectangle.Width - innerRectangle.Width) / 2.0);
		//    }
		//    else
		//    {
		//        innerRectangle.Height = (int)((double)outerRectangle.Width / innerRatio);
		//        innerRectangle.Y = (int)((outerRectangle.Height - innerRectangle.Height) / 2.0);
		//    }
		//    return innerRectangle;
		//}

		protected virtual void PaintBlackBands(Graphics g)
		{
			if (this.videoRenderer != null)
			{
				//Trace.WriteLineIf(trace.TraceInfo, "PaintBlackBands()");

				IVMRWindowlessControl9 vmrWindowlessControl9 = this.videoRenderer as IVMRWindowlessControl9;
				Rectangle outerRectangle = this.hostingControl.ClientRectangle;
				DsRect innerDsRect = new DsRect();
				vmrWindowlessControl9.GetVideoPosition(null, innerDsRect);
				Rectangle innerRectangle = innerDsRect.ToRectangle();

				//Trace.WriteLineIf(trace.TraceVerbose, string.Format(("\tvideoRenderer.GetVideoPosition({0})"), innerRectangle.ToString()));
				//Trace.WriteLineIf(trace.TraceVerbose, string.Format(("\thostingControl.ClientRectangle({0})"), outerRectangle.ToString()));

				ArrayList alRectangles = new ArrayList();

				if (innerRectangle.Top > outerRectangle.Top)
					alRectangles.Add(new Rectangle(outerRectangle.Left, outerRectangle.Top, outerRectangle.Width - 1, innerRectangle.Top - 1));

				if (innerRectangle.Bottom < outerRectangle.Bottom)
					alRectangles.Add(new Rectangle(outerRectangle.Left, innerRectangle.Bottom, outerRectangle.Width - 1, outerRectangle.Height - (innerRectangle.Bottom + 1)));

				if (innerRectangle.Left > outerRectangle.Left)
				{
					Rectangle rectangleLeft = new Rectangle(outerRectangle.Left, innerRectangle.Top, innerRectangle.Left - 1, innerRectangle.Height - 1);
					rectangleLeft.Intersect(outerRectangle);
					alRectangles.Add(rectangleLeft);
				}

				if (innerRectangle.Right < outerRectangle.Right)
				{
					Rectangle rectangleLeft = new Rectangle(innerRectangle.Right, innerRectangle.Top, outerRectangle.Width - (innerRectangle.Right + 1), innerRectangle.Height - 1);
					rectangleLeft.Intersect(outerRectangle);
					alRectangles.Add(rectangleLeft);
				}

				if (alRectangles.Count > 0)
				{
					g.FillRectangles(new SolidBrush(Settings.VideoBackgroundColor), (Rectangle[])alRectangles.ToArray(typeof(Rectangle)));
					g.DrawRectangles(new System.Drawing.Pen(Settings.VideoBackgroundColor), (Rectangle[])alRectangles.ToArray(typeof(Rectangle)));
				}
			}
		}

		public virtual void VideoRefresh()
		{
			if (this.videoRenderer != null)
			{
				//Trace.WriteLineIf(trace.TraceInfo, "VideoRefresh()");

				if (!useWPF)
				{
					VideoResizer(this.videoZoomMode, this.videoKeepAspectRatio, this.videoOffset, this.videoZoom, this.videoAspectRatioFactor);

					Graphics g = this.hostingControl.CreateGraphics();
					IntPtr hdc = g.GetHdc();
					try
					{
						int hr = (this.videoRenderer as IVMRWindowlessControl9).RepaintVideo(this.hostingControl.Handle, hdc);
					}
					catch { }
					g.ReleaseHdc(hdc);

					PaintBlackBands(g);

					g.Dispose();
				}
			}
		}

		// Idea from Gabest (http://www.gabest.org) and modify by me
		public virtual void VideoResizer(VideoSizeMode videoZoomMode, bool keepAspectRatio, PointF offset, double zoom, double aspectRatioFactor)
		{
			Trace.WriteLineIf(trace.TraceInfo, "VideoResizer(...)");
			int hr = 0;
			
			Rectangle windowRect = this.hostingControl.ClientRectangle;
			currentVideoTargetRectangle = windowRect;
			currentVideoSourceSize = new Size();

			if (videoZoomMode != VideoSizeMode.StretchToWindow)
			{
				FilterState filterState = GetGraphState();
				if (filterState == FilterState.Paused || filterState == FilterState.Running)
				{
					int arX, arY;
					int arX2 = 0, arY2 = 0;

					hr = (this.videoRenderer as IVMRWindowlessControl9).GetNativeVideoSize(out arX, out arY, out arX2, out arY2);
					if (hr >= 0 && arY > 0)
					{
						//DsError.ThrowExceptionForHR(hr);
						//Trace.WriteLineIf(trace.TraceVerbose, string.Format("\tGetNativeVideoSize(width: {0}, height: {1}, arX {2}, arY: {3}", arX, arY, arX2, arY2));

						if (arX2 > 0 && arY2 > 0)
						{
							arX = arX2;
							arY = arY2;
						}

						currentVideoSourceSize.Width = arX;
						currentVideoSourceSize.Height = arY;

						Size windowSize = windowRect.Size;

						double newAspectRation = aspectRatioFactor * (double)arX / (double)arY * (this.useVideo169Mode ? 3.0 / 4.0 : 1.0);
						int height = windowSize.Height;
						int width = (int)((double)height * newAspectRation);

						if (videoZoomMode == VideoSizeMode.FromInside || videoZoomMode == VideoSizeMode.FromOutside)
						{
							if (videoZoomMode == VideoSizeMode.FromInside && width > windowSize.Width
							|| videoZoomMode == VideoSizeMode.FromOutside && width < windowSize.Width)
							{
								width = windowSize.Width;
								height = (int)((double)width / newAspectRation);
							}
						}

						Size size = new Size((int)(zoom * width), (int)(zoom * height));

						Point pos = new Point(
							(int)(offset.X * (windowRect.Width * 3 - size.Width) - windowRect.Width),
							(int)(offset.Y * (windowRect.Height * 3 - size.Height) - windowRect.Height));

						//Point pos = new Point(
						//    (int)(offset.X * (windowRect.Width - size.Width)),
						//    (int)(offset.Y * (windowRect.Height - size.Height)));

						currentVideoTargetRectangle = new Rectangle(pos, size);
					}
				}
			}

			hr = (this.videoRenderer as IVMRWindowlessControl9).SetVideoPosition(null, DsRect.FromRectangle(currentVideoTargetRectangle));
			//Trace.WriteLineIf(trace.TraceVerbose, string.Format(("\tPos {0:F2} {1:F2}, Zoom {2:F2}, ARF {4:F2}, AR {4:F2}"), offset.X, offset.Y, zoom, aspectRatioFactor, (float)videoTargetRect.Width / videoTargetRect.Height));
			Trace.WriteLineIf(trace.TraceVerbose, string.Format(("\tvideoRenderer.SetVideoPosition({0})"), currentVideoTargetRectangle.ToString()));
		}

		protected virtual void OnResizeMoveHandler(object sender, EventArgs e)
		{
			if (this.videoRenderer != null)
			{
				//Trace.WriteLineIf(trace.TraceInfo, "OnResizeMoveHandler()");

				VideoResizer(this.videoZoomMode, this.videoKeepAspectRatio, this.videoOffset, this.videoZoom, this.videoAspectRatioFactor);

				Graphics g = this.hostingControl.CreateGraphics();
				PaintBlackBands(g);
				g.Dispose();
			}
		}

		protected virtual bool OnPaintBackground(object sender, PaintEventArgs e)
		{
			// The OnPaintBackground is call a lot often than OnPaintHandler!!!!
			if (this.videoRenderer != null)
			{
				//Trace.WriteLineIf(trace.TraceInfo, "OnPaintBackground()");
				return false;
			}
			return true;
		}

		void OnPaintHandler(object sender, PaintEventArgs e)
		{
			if (this.videoRenderer != null)
			{
				//Trace.WriteLineIf(trace.TraceInfo, "OnPaintHandler()");

				IntPtr hdc = e.Graphics.GetHdc();
				try
				{
					int hr = (this.videoRenderer as IVMRWindowlessControl9).RepaintVideo(this.hostingControl.Handle, hdc);
				}
				catch { }

				e.Graphics.ReleaseHdc(hdc);

				PaintBlackBands(e.Graphics);
			}
		}

		protected virtual void OnDisplayChangedHandler(object sender, EventArgs e)
		{
			if (this.videoRenderer != null)
			{
				//Trace.WriteLineIf(trace.TraceInfo, "OnDisplayChangedHandler()");

				int hr = (this.videoRenderer as IVMRWindowlessControl9).DisplayModeChanged();
			}
		}
	}
}
